-- GDL

local version = 18

-- http://en.wikipedia.org/wiki/Youtube#Quality_and_codecs
local qDesc = {
	    [5] = '  224p FLV'
	,   [6] = '  270p FLV'
	,  [34] = '  360p FLV'
	,  [35] = '  480p FLV'
	,  [18] = '  360p MP4'
	,  [22] = '  720p MP4'
	,  [37] = '1080p MP4'
	,  [38] = '2304p MP4'
	,  [43] = '  360p WebM'
	,  [44] = '  480p WebM'
	,  [45] = '  720p WebM'
	,  [46] = '1080p WebM'
	,  [13] = '3GP 176x144'
	,  [17] = '3GP 176x144'
	,  [36] = '3GP MP4'

	,  [83] = '  240p MP4 3D'
	,  [82] = '  360p MP4 3D'
	,  [85] = '  520p MP4 3D'
	,  [84] = '  720p MP4 3D'
	, [100] = '  360p WebM 3D'
	, [101] = '  480p WebM 3D'
	, [102] = '  720p WebM 3D'

    , [138] = 'mp4 no sound'
    , [137] = 'mp4 no sound'
    , [136] = 'mp4 no sound'
    , [135] = 'mp4 no sound'
    , [134] = 'mp4 no sound'
    , [133] = 'mp4 no sound'
    , [160] = 'mp4 no sound'
    , [264] = 'mp4 no sound'

	, [139] = 'audio mp4   48kb'
	, [140] = 'audio mp4 128kb'
	, [141] = 'audio mp4 256kb'
}




local Exts = {}
local function makeExts(tbl, ext, ...) for k,v in ipairs(arg) do tbl[v] = ext end end
makeExts(Exts, '.flv', 5,6,34,35)
makeExts(Exts, '.mp4', 18,22,37,38,82,83,84,85,133,134,135,136,137,160)
makeExts(Exts, '.webm', 43,44,45,46,100,101,102)
makeExts(Exts, '.webm', 171,172) -- audio
makeExts(Exts, '.m4a', 139,140,141) -- audio
makeExts(Exts, '.3gp', 13,17,36)


local conf = tableCombine({
		  interactive = false
		, useFallbackHost = true
		, skip3D = true
		, adaptiveFmts = false
		, nameTemplate = '%title%'
	}, Settings[script.name] or {})
conf.qualityList = conf.qualityList or {38, 37, 22, 45, 44, 35, 43, 34, 18, 5}
conf.subtitles = tableCombine({translateSrc='en'}, conf.subtitles or {})
local sconf = conf.subtitles
sconf.enable = sconf.xml or sconf.srt or sconf.sbv


local function parse(str, spaces)
	local t = {}
	for k, v in string.gmatch(str, '([^=&]+)=([^&]*)') do
		if spaces then
			--if k == 'title' or k == 'reason' then
				v = string.gsub(v, '%+', ' ')
			--end
		end
		t[k] = tonumber(v) or unescape(v)
	end
	return t
end

local function parseDataCommas(data)
	for _, n in ipairs({'url_encoded_fmt_stream_map', 'adaptive_fmts'}) do
		if data[n] then
			local tmp = {}
			for t in data[n]:gmatch('[^,]+') do
				table.insert(tmp, parse(t, true))
			end
			data[n] = tmp
		end
	end
end

local function parseData(str)
	local data = parse(str, true)
	if data.stream then data.stream = parse(str.stream) end
	parseDataCommas(data)
	return data
end

local function sd1586(s)	-- http://s.ytimg.com/yts/swfbin/watch_as3-vfl-o51IB.swf
	local s1 = {s:reverse():byte(1,#s)}
	s1[#s1-52], s1[#s1] = s1[#s1], s1[#s1-52]
	s1[25], s1[4] = s1[4], s1[25]
	return string.char(unpack(s1)):sub(4,-4)
end

local function blocked(data)
	return data.errorcode
		and data.errorcode == 150
		and data.reason
		and (
			data.reason:find('^Embedding disabled')
			or data.reason:find('restricted from playback on certain sites')
			or data.reason:find('blocked it from display on this website')
		)
end

--local countryBlocked = function(str)
--	return str:find('errorcode=150') and (
--		str:find('not+made+this+video+available+in+your+country',1,true)
--		or str:find('blocked+it+in+your+country',1,true)
--	)
--end
local countryBlocked = function(data) return data.status=="fail" and data.errorcode==150 and data.reason:find('country') end

local hosts = {}

--local function head(url, fb)
--	if usd.progstatus() == 0 then return end
--	local tmp = usd.head(url)
--	local h = getHost(usd.getlocation())
--	hosts[h]=h
--	return tmp
--end


function GDL(origUrl) -- GetDirectLink
	usd.setstatus(script.name)
	logpref = script.name .. '_GDL_v'..version
	local page, badpage, dirUrl

	local qList = conf.qualityList
	local dirUrl

	usd.setcookie('PREF=hl=en;')

	script.tmp={}

	--hosts = {}
	--hostsFB = {}
	--hostsBad = {}

	--   
	local fmt = string.match(origUrl, '[?&]fmt=(%d+)')
	if fmt then
		fmt = tonumber(fmt)
	end

	origUrl = unescape(origUrl)
	local id = string.match(origUrl, '[/?&]v[/=]([%w_-]+)')
		or string.match(origUrl, '/embed/([%w_-]+)')
		or string.match(origUrl, 'http://youtu.be/([%w_-]+)')
	if not id then
		usd.setretresult('%ID not found')
		return
	end
	origUrl = 'http://www.youtube.com/watch?v=' .. id
	--usd.modquery(origUrl  .. (fmt and '&fmt='..fmt or ''))
	usd.setreferer(origUrl)

	debugLog(serialize(conf,nil,nil,true))

	local data

	for _, v in ipairs({'', 'http://www.google.com/', 'http://www.youtube.com/'}) do
		usd.setreferer(v)
		--page = usd.get('http://www.youtube.com/get_video_info?sts=1586&video_id='..id..'&hl=en&eurl='..v)
		local url = 'http://www.youtube.com/get_video_info?sts=1586&hl=en&el=detailpage&ps=default&gl=US&video_id='..id..'&eurl='..v
		page = usd.get(url)
		data = parseData(page)
		debugSaveInc(logpref..'.html', page..'\r\n', serialize(data,nil,true))
		if not blocked(data) then break end
	end


--reason=Embedding+disabled+by+request%3Cbr%2F%3E%3Cu%3E%3Ca+href%3D%27http%3A%2F%2Fwww.youtube.com%2Fwatch%3Fv%3DPVbwIJ7oHnU%26feature%3Dplayer_embedded%27+target%3D%27_blank%27%3EWatch+on+YouTube%3C%2Fa%3E%3C%2Fu%3E&errorcode=150&storyboard_spec=http%3A%2F%2Fi1.ytimg.com%2Fsb%2FPVbwIJ7oHnU%2Fstoryboard3_L%24L%2F%24N.jpg%7C48%2327%23100%2310%2310%230%23default%23C0nRCH_zU8VHDt_DF3ufc5f13-g%7C62%2345%23443%2310%2310%2310000%23M%24M%23Seiqhw936iQkh5kRmfE46W9GsO8%7C124%2390%23443%235%235%2310000%23M%24M%238p2FYy2cpSAn8GkaycBSnBndpLE&status=fail&loudness=-31.888999939&errordetail=0
-- VO275GDVjvY: status="fail",errorcode=150,reason="      NBC Universal,         ."
-- JA4W1EwtyRM: status="fail",errorcode=100,reason="    ."

-- http://www.youtube.com/watch?v=F4P3j8dh2ro

	--if data.status=='fail'
	--	and data.errorcode == 150
	--	and ( data.reason:find('^Embedding disabled')
	--		or data.reason:find('restricted from playback on certain sites')
	--		or data.reason:find('blocked it from display on this website')
	--	)
	if blocked(data) then
		data.reason = data.reason:gsub('<[^>]+>',' ')
		usd.savetolog('Error: '..data.reason)
		usd.setreferer('')
		page = usd.get(origUrl..'&hl=en')
		debugSaveInc(logpref..'.html', page)
		data = page:match('"args": {.-}')
		if data then
			data = parseJsObject(data)
		else
			local tmp = page:match('<param name="flashvars" value="([^"]+)"') or ''	-- ie6
			tmp = entityDecode(tmp)
			data = parse(tmp)
		end

		if not ( data and next(data) ) then
			local err = page:match('<div id="verify%-actions">%s*<p>(.-)</p>')
				or page:match('Sign in to confirm your age')
				or 'plugindefect'
			err = utf8ToAnsi(err)
			return usd.setretresult('%'..err, 10)
		end

		parseDataCommas(data)
		debugSaveInc(logpref..'.html', serialize(data, nil, true), nil, false)

	elseif data.status=='fail' then
		local r = utf8ToAnsi(data.reason)
		r = r:gsub('<[^>]+>',' ')
		usd.savetolog(r)
		if data.errorcode==100
			or data.errorcode==2
			or data.reason=='This video is not available.'
			or conf.delCountryBlocked and countryBlocked(data)
			then
			usd.setretresult('%deleted%')
		else
			usd.setretresult('%'..r, 10)
		end
		return
	end

	if ( page:find('rtmpe%%3Dyes') or page:find('rtmpe%%253Dyes') )
		and (page:find('conn=rtmpe') or page:find('conn%%3Drtmpe') )
		then
		usd.setretresult('%RTMPE is not supported')
		return
	end

	if ( not data.url_encoded_fmt_stream_map or #data.url_encoded_fmt_stream_map==0 ) and data.hlsvp then
		usd.setretresult('%hlsvp is not supported')
		return
	end

	if not data.url_encoded_fmt_stream_map then
		usd.setretresult('%Download link not found!')
		return
	end


	local title = conf.nameTemplate
	title = utf8ToAnsi( type(title)=='function' and title(data) or title:gsub('%%([%w_-]+)%%',data) ):gsub('\151','-')
	newName = filterName( title ):gsub('[%s_]+','_'):gsub('_?%-_?','-'):gsub('^%-', '_-')

	if data.ttsurl and sconf.enable then
		--local tmp = usd.get(data.ttsurl..'&type=list&tlangs=1'..(sconf.asrs and '&asrs=1' or ''))
		local tmp = usd.get(data.ttsurl..'&type=list&tlangs=1&asrs=1')
		local subtitles = {asrs={},tracks={},targets={}}

--<track id="1" name="English v0.1" lang_code="en" lang_original="English" lang_translated="English" lang_default="true" cantran="true"/>
		for t in tmp:gmatch('<track[^>]+>') do
			--local id = t:match('id="(%d+)"')
			local name = t:match('name="([^"]-)"')
			local code = t:match('lang_code="([^"]+)"')
			local asr = t:find('kind="asr"')
			--if id and name and code then subtitles.tracks[tonumber(id)] = {name=name,code=code} end
			if name and code then subtitles[asr and 'asrs' or 'tracks'][code] = name end
		end

--<target id="44" lang_code="sw" lang_original="Kiswahili" lang_translated="Swahili"/>
		for t in tmp:gmatch('<target[^>]+>') do
			--local id = t:match('id="(%d+)"')
			local code = t:match('lang_code="([^"]+)"')
			local lang = t:match('lang_translated="([^"]+)"')
			if lang and code then subtitles.targets[code] = lang end
		end

		debugSaveInc(logpref..'.html', tmp..'\r\n', serialize(subtitles,nil,true))

		tmp = false
		for _,l in ipairs(sconf.langs) do
			if subtitles.tracks[l] then
				tmp = data.ttsurl..'&name='..subtitles.tracks[l]..'&lang='..l
				break
			end
		end

		if not tmp and sconf.asrs then
			for _,l in ipairs(sconf.langs) do
				if subtitles.asrs[l] then
					tmp = data.ttsurl..'&kind=asr&name='..subtitles.asrs[l]..'&lang='..l
					break
				end
			end
		end

		if not tmp and sconf.translate then
			local tfrom = subtitles.tracks[sconf.translateSrc] and sconf.translateSrc
				or sconf.asrs and subtitles.asrs[sconf.translateSrc] and sconf.translateSrc
				or next(subtitles.tracks)
				or sconf.asrs and next(subtitles.asrs)
			if tfrom then
				tmp = data.ttsurl..'&name='..(subtitles.tracks[tfrom] or subtitles.asrs[tfrom]..'&kind=asr')..'&lang='..tfrom..'&tlang='..sconf.translate
			end
		end

		if tmp then
			script.tmp.sfile = usd.workdir..id
			if sconf.xml then usd.get(tmp..'|'..script.tmp.sfile..'.xml') end
			if sconf.srt then usd.get(tmp..'&fmt=srt|'..script.tmp.sfile..'.srt') end
			if sconf.sbv then usd.get(tmp..'&fmt=sbv|'..script.tmp.sfile..'.sbv') end
		end
	end




	local res = {}	-- 
	for k, v in string.gmatch(','..data.fmt_list, ',(%d+)/([%dx]+)') do
		res[tonumber(k)] = v
	end

	local urls = {} --         
	local info = {} --  

	function parseStreams(streamData)
		for _, stream in ipairs(streamData) do
			if usd.progstatus() == 0 then  debugLog('USD stopped');  return  end
			if stream.stream then stream.stream = parse(stream.stream) end
			stream.itag = tonumber(stream.itag)
			--if stream.fallback_host then hostsFB[stream.fallback_host]=stream.fallback_host end
		end

		local fb = conf.useFallbackHost
		local fbh


		for _, stream in ipairs(streamData) do
			local signature = not stream.url:find('&signature=')
				and (stream.sig and '&signature='..stream.sig or stream.s and '&signature='..sd1586(stream.s))
				or ''
			local video = stream.url .. signature

			if fb then
				fbh = fb and ( type(fb)=='string' and fb or stream.fallback_host )
				if fbh then
					video = video:gsub('^http://.-/','http://'..fbh..'/',1)
				end
			end

			if not ( conf.skip3D and ((qDesc[stream.itag] or ''):find(' 3D$') or stream.stereo3d) ) then

				if usd.progstatus() == 0 then return debugLog('USD stopped') end
				local sDesc = qDesc[stream.itag] or stream.type or '<unknown>'
				if conf.interactive and not fmt then
					--debugSaveInc(logpref..'.html', usd.head(video), ('%s\r\n%s\r\n%s\r\n%s'):format(
					debugSaveInc(logpref..'.html', urlHead(video, origUrl), ('%s\r\n%s\r\n%s\r\n%s'):format(
						origUrl
						, sDesc
						, video
						, serialize(stream, nil, true) )
					)
					urls[stream.itag] = usd.getlocation()
				else
					urls[stream.itag] = video
				end
				local size = usd.getheaders():match('HTTP/%d%.%d 200 OK') and usd.getheaders():match('Content%-Length: (%d+)')
				size = size and string.format('%0.2f', size / 1024 / 1024) or ' ??????'

				if conf.interactive and not fmt then
					table.insert(info
						, string.format(
							'%s: %s MB   ~%s   %s'
							, pad(stream.itag,3, '  ')
							, pad(size, 7, '  ')
							, pad(res[stream.itag] or stream.size, 9, '  ')
							, sDesc
						)
					)
				end
			end
		end

	end

	parseStreams(data.url_encoded_fmt_stream_map)
	if conf.adaptiveFmts and data.adaptive_fmts then parseStreams(data.adaptive_fmts) end

	--debugSaveInc(logpref..'.html', ('hosts\r\n%s\r\n\r\nhostsFB\r\n%s\r\n\r\nhostsBad\r\n%s'):format(serialize(tableGetKeys(hosts),nil,true),serialize(tableGetKeys(hostsFB),nil,true),serialize(tableGetKeys(hostsBad),nil,true))	, nil, false)

	--debugSaveInc(logpref..'.html', serialize(tableGetKeys(hosts),nil,true), nil, false )

	if fmt and urls[fmt] then
		--debugSaveInc(logpref..'.html', usd.head(urls[fmt]), ('%s\r\n%s\r\n%s'):format(origUrl, fmt, qDesc[fmt] or '<unknown>') )
		debugSaveInc(logpref..'.html', urlHead(urls[fmt], origUrl), ('%s\r\n%s\r\n%s'):format(origUrl, fmt, qDesc[fmt] or '<unknown>') )
		dirUrl = usd.getlocation() .. '&title=' .. escapeUC(newName)
		dirUrl = addNewName(dirUrl, newName)
		usd.setretresult( dirUrl, 0, getResumeFlag() )
		return
	end

	if not next(urls) and data.url_encoded_fmt_stream_map[1].s then
		usd.setretresult('%Unsupported')
		return
	end

	if conf.interactive then
		local duration = math.floor(data.length_seconds / 60).. ':' .. pad(data.length_seconds % 60, 2)
		local dlg = {
			{duration, title}
			, {'__________________________'}
			, {'Select quality:'}
			, {{type='select', name='quality', info }}
		}
		dlg = Dialog.new(dlg, script.name)
		if usd.progstatus() == 0 then
			debugLog('USD stopped')
			return
		end
		local q = dlg:show()
		if q == '' then debugLog('Esc') return end		-- esc
		q = q:match('quality=([%d]+)') + 1	--     0
		q = tonumber(info[tonumber(q)]:match('^%s*(%d+)'))
		fmt = q
		usd.modquery(origUrl .. '&fmt='..fmt)
		dirUrl = urls[q] .. '&title=' .. escapeUC(newName)
		dirUrl = addNewName(dirUrl, newName)
		--usd.setretresult( dirUrl, 0, resume )
		usd.setretresult( dirUrl, 0, 0 )
		return
	end

	if table.maxn(urls) == 0 then
		usd.setretresult('%Download link not found!')
		return
	end
	for i,v in ipairs(qList) do
		if urls[v] then
			dirUrl = urls[v]
			fmt = v
			debugLog('GDL youtube: url index: '..v)
			break
		end
	end
	if dirUrl == nil then
		fmt = table.maxn(urls)
		dirUrl = urls[fmt]
		debugLog('GDL youtube: url index not found; selected max number: '..fmt)
	end

	usd.modquery(origUrl .. '&fmt='..fmt)
	debugSaveInc(logpref..'.html', urlHead(dirUrl, origUrl), origUrl..'\r\n'..fmt)
	dirUrl = usd.getlocation() .. '&title=' .. escapeUC(newName)
	dirUrl = addNewName(dirUrl, newName)
	--return usd.setretresult(dirUrl, 0, resume)
	return usd.setretresult(dirUrl, 0, 0)
end

function OS(origUrl, dirUrl, fullName)
	logpref = script.name .. '_OS_v'..version

	local fmt = tonumber(origUrl:match('&fmt=(%d+)') or '')
	local fmtExt = Exts[fmt]
	local fileExt = (fullName:match('%.[^\\/.]+$') or ''):lower()

	local newName = fullName
	if fmtExt ~= fileExt then
		newName = fileRename(fullName, nil, fmtExt)
	end

	if script.tmp.sfile then
		for _,v in ipairs({'.xml','.srt','.sbv'}) do
			local f=script.tmp.sfile..v
			if (fileSize(f) or 0)>24 then
				local _, err = os.rename(f, newName..v)
				if err then usd.savetolog(err) end
			end
		end
	end

	usd.setretresult(newName)
end

regDirSuccess('youtube.com/', GDL, OS)
regDirSuccess('//youtu.be/', GDL, OS)

